## 
package POE::Jukebox::DBFilePlaylist;
use strict;
use MLDBM qw(DB_File Storable);
use Audio::Mixer;

use Data::Dumper;

use POE qw(Driver::SysRW Wheel::ReadWrite Filter::Stream);

sub new
{
	my $self = shift;
	my @list = @_;
	
	POE::Session->create(
		args => [ @list ],
		inline_states => {
			_start => \&_start,
			_stop => \&_stop,
			songcomplete => \&playlist_song_complete,
			next => \&playlist_next,
			prev => \&playlist_prev,
			pause => \&playlist_pause,
			volume => \&volume,
			mode => \&playlist_mode,
			current => \&playlist_current,
			send_current_info => \&send_current_info,	
			set_fraction_complete => \&set_fraction_complete,	
		},
	);
}

sub _start
{
	my ($songfile,$albumfile,$artistfile,$genrefile) = @_[ARG0,ARG1,ARG2,ARG3];
	
	print "Starting playlist\n";
	
	$_[KERNEL]->alias_set( 'playlist' );

	print "Tying to DB files $songfile, $albumfile\n";
	tie(%{$_[HEAP]->{song_database}}, 'MLDBM',$songfile) or die $!;
	tie(%{$_[HEAP]->{album_database}}, 'MLDBM',$albumfile) or die $!;
	tie(%{$_[HEAP]->{artist_database}}, 'MLDBM',$artistfile) or die $!;
	tie(%{$_[HEAP]->{genre_database}}, 'MLDBM',$genrefile) or die $!;

	## Set up internal storage.
	$_[HEAP]->{play_mode} = 'allrandom';
	$_[HEAP]->{play_status} = 'stopped';
	$_[HEAP]->{song_fraction_complete} = 0;
	$_[HEAP]->{song_position} = 0;

	## Start up the playlist
	print "Generating Random Playlist\n";	
	$_[HEAP]->{playlist}->{all} = {
		playlist => [ keys %{$_[HEAP]->{song_database}} ],
		num_songs => scalar(keys %{$_[HEAP]->{song_database}}),
		current_idx => 0
	};
	fisher_yates_shuffle($_[HEAP]->{playlist}->{all}->{playlist});
	$_[HEAP]->{playlist}->{current} = $_[HEAP]->{playlist}->{all};
	
	print "Started playlist\n";

	## Set up signals.
	## TODO: Write better control method!
	$_[KERNEL]->sig('USR2' => 'next');
	$_[KERNEL]->sig('USR1' => 'prev');
	
	## Get things moving.
	$_[KERNEL]->yield('songcomplete');
}

sub _stop
{
	undef $_[HEAP]->{playlist}->{current}; # to prevent loops.
	delete $_[HEAP]->{self};
	untie(%{$_[HEAP]->{song_database}});
	untie(%{$_[HEAP]->{album_database}});
	
	print "Stopped playlist\n";
}

sub post_current_song
{
	my %db = %{$_[HEAP]->{song_database}};
	my $list = $_[HEAP]->{playlist}->{current};
	my $song = $db{$list->{playlist}->[$list->{current_idx}]};
	$_[HEAP]->{current_song} = $song;
	$_[HEAP]->{song_fraction_complete} = 0;
	$_[HEAP]->{song_position} = 0;
	print "$song->{track_title} by $song->{track_artist} (from $song->{album_title}) is up next\n";
	print "Number $list->{current_idx} of ", scalar(@{$list->{playlist}}), "\n";
	$_[HEAP]->{play_status} = 'playing';
	$_[KERNEL]->post('player','song_new', $song->{filename});
	$_[KERNEL]->post('clientmanager','write_category', 'update', current_info(@_));
}

sub playlist_song_complete
{
	my $playlist = $_[HEAP]->{playlist}->{current};
	print "Moving to next song\n";
	if(($playlist->{current_idx} + 1) < $playlist->{num_songs}) {
		$playlist->{current_idx}++;
	} else {
		# Wrap around if we try to go off the end.
		$playlist->{current_idx} = 0;
	}
	post_current_song(@_);
}


sub playlist_next
{
	my $session = $_[ARG0];
	my $playlist = $_[HEAP]->{playlist}->{current};
	print "Skipping to next song\n";
	if(($playlist->{current_idx} + 1) < $playlist->{num_songs}) {
		$playlist->{current_idx}++;
	} else {
		# Wrap around if we try to go off the end.
		$playlist->{current_idx} = 0;
	}
	$_[KERNEL]->post('player','song_stop');
	post_current_song(@_);
	$_[KERNEL]->post($session,'out_to_client','OK' );
}

sub playlist_pause
{
	my $session = $_[ARG0];
	my $playlist = $_[HEAP]->{playlist}->{current};
	$_[KERNEL]->post('player','song_pause');
	
	if($_[HEAP]->{play_status} eq 'playing') {
		print "Pausing\n";
		$_[HEAP]->{play_status} = 'paused';
	} elsif ($_[HEAP]->{play_status} eq 'paused') {
		print "Unpausing\n";
		$_[HEAP]->{play_status} = 'playing';
	}
	$_[KERNEL]->post('clientmanager','write_category', 'update', current_info(@_));
}

sub playlist_prev
{
	my $session = $_[ARG0];
	my $playlist = $_[HEAP]->{playlist}->{current};
	print "Skipping to previous song\n";
	if(($playlist->{current_idx} - 1) > 0) {
		$playlist->{current_idx}--;
	} else {
		# Wrap around if we try to go off the end.
		$playlist->{current_idx} = ($playlist->{num_songs} - 1);
	}
	$_[KERNEL]->post('player','song_stop');
	post_current_song(@_);
	$_[KERNEL]->post($session,'out_to_client','OK' );
}

sub send_current_info
{
	$_[KERNEL]->post('clientmanager','write_category', 'update',current_info(@_));
}

sub playlist_current
{
	my $session = $_[ARG0];
	$_[KERNEL]->post($session,'out_to_client', current_info(@_) );
}

sub set_fraction_complete
{
	my $new_fract = $_[ARG0];
	my $new_position;

	$new_position = $_[HEAP]->{current_song}->{length} * $new_fract;
	
	return if (abs($new_position - $_[HEAP]->{song_position}) < 1000);
	
	$_[HEAP]->{song_position} = int($new_position);
#	$_[HEAP]->{song_fraction_complete} = $new_fract;
#song_pct_complete: $_[HEAP]->{song_fraction_complete}
	my $data =<<EOF;
PROGRESS
song_position: $_[HEAP]->{song_position}
OK
EOF
	$_[KERNEL]->post('clientmanager','write_category','progress',$data);
}

sub current_info
{
	my $song = $_[HEAP]->{current_song};
	my ($left_vol,$right_vol) = Audio::Mixer::get_cval('vol');
	my $data =<<EOF;
UPDATE
track_title: $song->{track_title}
track_artist: $song->{track_artist}
album_title: $song->{album_title}
album_uuid: $song->{album_uuid}
track_number: $song->{number}
album_genre: $song->{genre}
length: $song->{length}
song_pct_complete: $_[HEAP]->{song_fraction_complete}
song_position: $_[HEAP]->{song_position}
play_status: $_[HEAP]->{play_status}
play_mode: $_[HEAP]->{play_mode}
left_volume: $left_vol
right_volume: $right_vol
OK
EOF
	chomp($data);
	return $data;
}

## TODO: implement a list_play_modes that will return what play 
## modes this playlist module supports.
## TODO: move this to a command manager
sub playlist_mode
{
	my $session = $_[ARG0];
	my $mode = $_[ARG1];
	if(!defined($mode)) {
		$_[KERNEL]->post($session,'out_to_client',"ERR No mode given");
		return;
	}
	# else, check the mode, and mangle appropriately
	if($mode eq 'album') {
		my %song_db = %{$_[HEAP]->{song_database}};
		my %album_db = %{$_[HEAP]->{album_database}};
		my $list = $_[HEAP]->{playlist}->{current};
		my $song = $song_db{$list->{playlist}->[$list->{current_idx}]};
	
		my $album = $album_db{$song->{album_uuid}};
		# Eww.  This is ugly, but it's cheaper than doing a Schwartzian
		# Transform.
		my @playlist;
		my $num_songs = scalar(keys %{$album->{tracks}});
		for(my $i=1; $i <= $num_songs; $i++) {
			##print "Adding $album->{tracks}->{$i} to playlist\n";
			push @playlist, $album->{tracks}->{$i} if exists($album->{tracks}->{$i});
		}
		delete($_[HEAP]->{playlist}->{thisalbum});
		$_[HEAP]->{playlist}->{thisalbum} = {
			playlist => \@playlist,
			num_songs => $num_songs,
			current_idx => $song->{number}
		};
		$_[HEAP]->{playlist}->{current} = $_[HEAP]->{playlist}->{thisalbum};
		$_[HEAP]->{play_mode} = 'album';
		
		$_[KERNEL]->post($session,'out_to_client','OK');
		return;
	}
	if($mode eq 'artist') {
		my %song_db = %{$_[HEAP]->{song_database}};
		my %artist_db = %{$_[HEAP]->{artist_database}};
		my $list = $_[HEAP]->{playlist}->{current};
		my $song = $song_db{$list->{playlist}->[$list->{current_idx}]};
		print "Current artist: $song->{track_artist}\n";
		my $artist = $artist_db{$song->{track_artist}};
		# Eww.  This is ugly, but it's cheaper than doing a Schwartzian
		# Transform.
		my @playlist;
		my $num_songs = scalar(@{$artist});
		my $current_idx = -1;
		for(my $i=0; $i < $num_songs; $i++) {
			my $item = $artist->[$i];
			$current_idx = $i if ($item eq $song->{uuid});
			push @playlist, $item;
		}
		
		delete($_[HEAP]->{playlist}->{thisartist});
		$_[HEAP]->{playlist}->{thisartist} = {
			playlist => \@playlist,
			num_songs => $num_songs,
			current_idx => $current_idx == -1 ? 1 : $current_idx,
		};
		fisher_yates_shuffle($_[HEAP]->{playlist}->{thisartist}->{playlist});
		$_[HEAP]->{playlist}->{current} = $_[HEAP]->{playlist}->{thisartist};
		$_[HEAP]->{play_mode} = 'artist';
		
		$_[KERNEL]->post($session,'out_to_client','OK');
		return;
	}
	if($mode eq 'genre') {
		my %song_db = %{$_[HEAP]->{song_database}};
		my %genre_db = %{$_[HEAP]->{genre_database}};
		my $list = $_[HEAP]->{playlist}->{current};
		my $song = $song_db{$list->{playlist}->[$list->{current_idx}]};
		print "Current genre: $song->{genre}\n";
		my $genre = $genre_db{$song->{genre}};
		# Eww.  This is ugly, but it's cheaper than doing a Schwartzian
		# Transform.
		my @playlist;
		my $num_songs = scalar(@{$genre});
		my $current_idx = -1;
		for(my $i=0; $i < $num_songs; $i++) {
			my $item = $genre->[$i];
			$current_idx = $i if ($item eq $song->{uuid});
			push @playlist, $item;
		}
		
		delete($_[HEAP]->{playlist}->{thisgenre});
		$_[HEAP]->{playlist}->{thisgenre} = {
			playlist => \@playlist,
			num_songs => $num_songs,
			current_idx => $current_idx == -1 ? 1 : $current_idx,
		};
		fisher_yates_shuffle($_[HEAP]->{playlist}->{thisgenre}->{playlist});
		$_[HEAP]->{playlist}->{current} = $_[HEAP]->{playlist}->{thisgenre};
		$_[HEAP]->{play_mode} = 'genre';
		
		$_[KERNEL]->post($session,'out_to_client','OK');
		return;
	}
	if($mode eq 'allrandom') {
		$_[HEAP]->{play_mode} = 'allrandom';
		$_[HEAP]->{playlist}->{current} = $_[HEAP]->{playlist}->{all};
		$_[KERNEL]->post($session,'out_to_client','OK');
		return;
	}
	$_[KERNEL]->post($session,'out_to_client','ERR Invalid mode given');
}

sub fisher_yates_shuffle {
    my $array = shift;
    my $i;
	
	print "Shuffling playlist\n";
	
    for ($i = @$array; $i--;) {
#        my $j = int rand ($i+1);
        my $j = int rand (scalar(@$array));
        next if $i == $j;
        @$array[$i,$j] = @$array[$j,$i]
    }
}

1;
