## Player backend for POE Jukebox
## This uses ogg123 as there's no good Ogg/Vorbis API in Perl (yet)
## It's either playing or stopped.  Features still to implement are:
##   - pause (just manually set ok_to_send = 0)
##   - resume (set ok_to_send to 1)
##   - seek forward/backwards (only for local files, not http streams)
use strict;

package POE::Jukebox::OggPlayer;

use POE qw(Driver::SysRW Wheel::ReadWrite Filter::Stream Filter::Block);

use Socket;
use IO::File;
use IO::Pipe;

sub new
{
	my $class = shift;
	my $self = bless({ }, $class);
	POE::Session->create(
		inline_states => {
			_start => \&_start,
			_stop => \&_stop,
			song_new => \&song_new,     ## Play new song 
			song_stop => \&song_stop,	## Stop current song
			read_chunk => \&read_chunk,
			wheel_throttle => \&wheel_throttle,
			wheel_resume => \&wheel_resume,
		},
	);

	print "Created player\n";
}

sub _start
{
	## Fork off the player
	$_[HEAP]->{player} = new IO::Pipe;
	$_[HEAP]->{player}->writer(qw(/usr/bin/ogg123 -q -));

	$_[KERNEL]->alias_set( 'player' );

	my $fh = new IO::File;
	
	$_[HEAP]->{filehandle} = $fh;
	$_[HEAP]->{chunk_size} = 1024;
	$_[HEAP]->{okay_to_send} = 1;

	## Set up wheel for i/o
	$_[HEAP]->{wheel} = POE::Wheel::ReadWrite->new(
		Handle => $_[HEAP]->{player},
		Driver => new POE::Driver::SysRW(BlockSize => $_[HEAP]->{chunk_size}),
		Filter => new POE::Filter::Stream,
		HighMark     => 8 * $_[HEAP]->{chunk_size},
		LowMark		=> 4 * $_[HEAP]->{chunk_size},
		HighEvent    => 'wheel_throttle',
		LowEvent     => 'wheel_resume',
	);
	
	print "Started player\n";
	$_[KERNEL]->post('playlist','songcomplete');
}

sub _stop
{
	delete $_[HEAP]->{self};
	print "Stopped player\n";
}

sub wheel_throttle
{
	#warn "Writer session ", $_[SESSION]->ID, " is throttled.\n";
	$_[HEAP]->{okay_to_send} = 0;
}

sub wheel_resume
{
	if (exists $_[HEAP]->{wheel}) {
		#warn "Writer session ", $_[SESSION]->ID, " is resuming.\n";
		$_[HEAP]->{okay_to_send} = 1;
		$_[KERNEL]->yield('read_chunk');
	}
}

sub read_chunk 
{
	my $chunk;
	my $fh = $_[HEAP]->{filehandle};

	if(!$_[HEAP]->{okay_to_send}) {
		return;
	}
	if(!$fh->opened()) {
		return;
	}
	my $len = $fh->sysread($chunk,$_[HEAP]->{chunk_size});
	if(!defined($len)) {
		warn "Read error: $!";
		$_[KERNEL]->post('playlist','songcomplete');
		return;
	}
	if($len > 0) {
#		print "Read $len (OK)\n";
		$_[HEAP]->{wheel}->put($chunk);
		$_[KERNEL]->yield('read_chunk');
	} else {
		print "Read hit end of file\n";
		$_[KERNEL]->post('playlist','songcomplete');
	}
}

sub song_new
{
	my $filename = $_[ARG0];

	my $fh = $_[HEAP]->{filehandle};

	## print "Trying to open song $filename\n";
	if($fh->opened) {
		$fh->close();
	}

	if(!$fh->open("<$filename")) {
		warn "Error opening file $filename: $!";
		return;
	}

	$_[KERNEL]->yield('read_chunk');
	
}

## Stop the current song from playing.
sub song_stop
{
	$_[HEAP]->{filehandle}->close();
}

1;
