## Player backend for POE Jukebox
## This uses ogg123 as there's no good Ogg/Vorbis API in Perl (yet)
## Features still to implement are:
##   - seek forward/backwards (only for local files, not http streams)
use strict;

package POE::Jukebox::OggVorbisPlayer;

use POE qw(Driver::SysRW Wheel::Run Filter::Stream Filter::Block);

use Ogg::Vorbis;
use Audio::OSS qw(:funcs :formats);
use Socket;
use IO::File;
use IO::Handle;
use Fcntl;

sub new
{
	my $class = shift;
	my $self = bless({ }, $class);
	POE::Session->create(
		inline_states => {
			_start => \&_start,
			_stop => \&_stop,
			song_new => \&song_new,     ## Play new song 
			song_stop => \&song_stop,	## Stop current song
			song_pause => \&song_pause,	## Pause current song
			read_chunk => \&read_chunk,
			wheel_throttle => \&wheel_throttle,
			wheel_resume => \&wheel_resume,
		},
	);

	print "Created player\n";
}

sub _start
{
	$_[KERNEL]->alias_set( 'player' );

	$_[HEAP]->{output} = new IO::Handle;
	sysopen $_[HEAP]->{output}, "/dev/dsp", O_WRONLY or die "open of /dev/dsp failed: $!";
	dsp_reset($_[HEAP]->{output});
	set_stereo($_[HEAP]->{output},1);
	set_fmt($_[HEAP]->{output},AFMT_S16_LE);
	set_sps($_[HEAP]->{output}, 44100);
	
	$_[HEAP]->{in_file} = new IO::File;
	$_[HEAP]->{buffer_size} = 2048;
	$_[HEAP]->{buffer} = 'x' x 4096; #$_[HEAP]->{buffer_size};
	$_[HEAP]->{okay_to_send} = 1;
	$_[HEAP]->{current_stream} = 0;
	$_[HEAP]->{oggplayer} = Ogg::Vorbis->new;
	$_[HEAP]->{host_big_endian_p} = Ogg::Vorbis::host_is_big_endian();

	$_[HEAP]->{wheel} = POE::Wheel::ReadWrite->new(
		Handle => $_[HEAP]->{output},
		Driver => new POE::Driver::SysRW(BlockSize => $_[HEAP]->{buffer_size}),
		Filter => new POE::Filter::Stream,
		HighMark     => 16 * $_[HEAP]->{buffer_size},
		LowMark     => 8 * $_[HEAP]->{buffer_size},
		HighEvent    => 'wheel_throttle',
		LowEvent     => 'wheel_resume',
	);

	print "Started player\n";
	$_[KERNEL]->post('playlist','songcomplete');
}

sub _stop
{
	delete $_[HEAP]->{oggplayer};
	delete $_[HEAP]->{self};	
	print "Stopped player\n";
}

sub wheel_throttle
{
#	warn "Writer session ", $_[SESSION]->ID, " is throttled.\n";
	$_[HEAP]->{okay_to_send} = 0;
}

sub wheel_resume
{
	if (exists $_[HEAP]->{wheel}) {
#		warn "Writer session ", $_[SESSION]->ID, " is resuming.\n";
		$_[HEAP]->{okay_to_send} = 1;
		$_[KERNEL]->yield('read_chunk');
	}
}

sub read_chunk 
{
	if(!$_[HEAP]->{okay_to_send}) {
		return;
	}
	my $bytes = $_[HEAP]->{oggplayer}->read(
			$_[HEAP]->{buffer},
			4096,#$_[HEAP]->{buffer_size},
			$_[HEAP]->{host_big_endian_p},
			2, 1, $_[HEAP]->{current_stream});
	if($bytes > 0) {
		# Successful read, play it on the device.
		$_[HEAP]->{wheel}->put($_[HEAP]->{buffer});
	} elsif($bytes < 0) {
		## Error
		warn "error in stream\n"
	} else {
		## End of stream
		$_[KERNEL]->post('playlist','songcomplete');
		return;
	}
	$_[KERNEL]->yield('read_chunk');
}

sub song_new
{
	my $filename = $_[ARG0];

	my $fh = $_[HEAP]->{in_file};
	my $ogg = $_[HEAP]->{oggplayer};

	print STDERR "Trying to open song $filename\n";
	if($fh->opened) {
		$fh->close();
	}

	if(!$fh->open("<$filename")) {
		warn "Error opening file $filename: $!";
		$_[KERNEL]->post('playlist','songcomplete');
		return;
	}

	if($ogg->open($fh) < 0) {
		warn "Error opening file $filename: $!";
		$_[KERNEL]->post('playlist','songcomplete');
		return;
	}
##	my $t_time = $ogg->time_total;
##	my $t_min = int($t_time / 60);
##	my $t_sec = $t_time % 60;

	$_[KERNEL]->yield('read_chunk');
	
}
## Stop the current song from playing.
sub song_stop
{
	$_[HEAP]->{in_file}->close();
}

## Pause/unpause the current song
sub song_pause
{
	$_[HEAP]->{okay_to_send} = !$_[HEAP]->{okay_to_send};
	if($_[HEAP]->{okay_to_send}) {
		$_[KERNEL]->yield('read_chunk');
	}
}

1;
