## Player backend for POE Jukebox
## This uses ogg123 as there's no good Ogg/Vorbis API in Perl (yet)
## Features still to implement are:
##   - seek forward/backwards (only for local files, not http streams)
use strict;

package POE::Jukebox::WheelRunOggPlayer;

use POE qw(Driver::SysRW Wheel::Run Filter::Stream Filter::Block);
use IO::File;

sub new
{
	my $class = shift;
	my $self = bless({ }, $class);
	POE::Session->create(
		inline_states => {
			_start => \&_start,
			_stop => \&_stop,
			_signal => \&signal_handler,
			song_new => \&song_new,     ## Play new song 
			song_stop => \&song_stop,	## Stop current song
			song_pause => \&song_pause,	## Pause current song
			read_chunk => \&read_chunk,
		},
	);

	print "Created player\n";
}

sub _start
{
	## Fork off the player
	$_[KERNEL]->alias_set( 'player' );

	my $fh = new IO::File;
	
	$_[HEAP]->{filehandle} = $fh;
	$_[HEAP]->{chunk_size} = 1024;
	$_[HEAP]->{okay_to_send} = 1;
	$_[HEAP]->{file_size} = 0;
	$_[HEAP]->{file_seen_length} = 0;

	## Set up wheel for i/o
	$_[HEAP]->{wheel} = POE::Wheel::Run->new(
		Program => [qw(/usr/bin/ogg123 -q -)],
		Filter => new POE::Filter::Stream,
		ErrorEvent => sub { die "Horrible $_[ARG0] error happened: $_[ARG2]" },
		StdinEvent  => 'read_chunk',
	);
		
	print "Started player\n";
	$_[KERNEL]->post('playlist','songcomplete');
}

sub _stop
{
	$_[HEAP]->{wheel}->kill();
	delete $_[HEAP]->{self};	
	print "Stopped player\n";
}

sub signal_handler
{
	my $signal = $_[ARG0];
	if($signal eq 'PIPE') {
		## Our ogg123 process is dead
		## Set up new wheel 
		
		$_[HEAP]->{wheel}->kill();
		delete $_[HEAP]->{wheel};
		
		$_[HEAP]->{wheel} = POE::Wheel::Run->new(
			Program => [qw(/usr/bin/ogg123 -q -)],
			Filter => new POE::Filter::Stream,
			ErrorEvent => sub { die "Horrible $_[ARG0] error happened: $_[ARG2]" },
			StdinEvent  => 'read_chunk',
		);
		print "Restarted dead player\n";
		
		$_[KERNEL]->post('playlist','songcomplete');
##	} else {
##		print "Player got unhandled signal $signal\n";
	}
}

sub read_chunk 
{
	my $fh = $_[HEAP]->{filehandle};
	my $pct_complete = 0;

	if(!$_[HEAP]->{okay_to_send}) {
		return;
	}
	
	if(!$fh->opened()) {
		return;
	}
	
	my $len = $fh->sysread($_[HEAP]->{chunk_buffer},$_[HEAP]->{chunk_size});
	if(!defined($len)) {
		warn "Read error: $!";
		$_[KERNEL]->post('playlist','songcomplete');
		return;
	}
	if($len > 0) {
		$_[HEAP]->{file_seen_length} += $len;
		$_[KERNEL]->post('playlist','set_fraction_complete',($_[HEAP]->{file_seen_length} / $_[HEAP]->{file_size}));
		$_[HEAP]->{wheel}->put($_[HEAP]->{chunk_buffer});
	} else {
		print STDERR "Read hit end of file\n";
		$_[KERNEL]->post('playlist','songcomplete');
	}
}

sub song_new
{
	my $filename = $_[ARG0];

	my $fh = $_[HEAP]->{filehandle};


	print STDERR "Trying to open song $filename\n";
	if($fh->opened) {
		$fh->close();
	}

	if(!$fh->open("<$filename")) {
		warn "Error opening file $filename: $!";
		return;
	}
	$_[HEAP]->{file_size} = -s $filename;
	$_[HEAP]->{file_seen_length} = 0;

	$_[KERNEL]->yield('read_chunk');
}

## Stop the current song from playing.
sub song_stop
{
	$_[HEAP]->{file_size} = 0;
	$_[HEAP]->{file_seen_length} = 0;
	$_[HEAP]->{filehandle}->close();
}

## Pause/unpause the current song
sub song_pause
{
	$_[HEAP]->{okay_to_send} = !$_[HEAP]->{okay_to_send};
	if($_[HEAP]->{okay_to_send}) {
		$_[KERNEL]->yield('read_chunk');
	}
}

1;
